from .settings_class import SettingClass
from .settings_page import SettingsPage
from .settings_element import SettingsElement

class SettingProperty:

    def __init__(self, setting_class, *args, **kwargs):
        """ Attributes of 'SettingProperty' """
        self.setting_class = setting_class
        self.args = args
        self.kwargs = kwargs
        self.se: SettingsElement | None = None

    def __set_name__(self, owner, name):
        self.kwargs['key'] = name
        self.public_name = name
        self.private_name = '_' + name
        owner: SettingClass
        self.se: SettingsElement = self.setting_class(*self.args, **self.kwargs)

        if 'elements' in owner.__dict__:
            owner.elements[name] = self.se
        else:
            owner.elements = {name: self.se}

    def __get__(self, obj, objtype=None):
        return getattr(obj, self.private_name, self.se.default)

    def __set__(self, obj, value):
        value = self.se._parse_value(value)
        self.se.to_validation(value)
        setattr(obj, self.private_name, value)



class FormNotFound(Exception):
    def __init__(self, form, **data):
        self.message = f"Form '{str(form)}' is not a SubClass from SettingsForm !"
        self.append_data = data
        super().__init__(self.message)

    def __str__(self):
        return f'[FlaskSettings] {self.message}'

class ValidationError(Exception):
    def __init__(self, message=None, **data):
        self.message = message or "Error on Validation"
        self.append_data = data
        super().__init__(self.message)

    def __str__(self):
        return f'[FlaskSettings] {self.message}'


class ParsingError(Exception):
    def __init__(self, message=None, **data):
        self.message = message or "Error on Parsing"
        self.append_data = data
        super().__init__(self.message)

    def __str__(self):
        return f'[FlaskSettings] {self.message}'


class PropertyPermissionError(Exception):
    def __init__(self, message=None, property_name=None, **data):
        self.message = message or f"PermissionError on Property ({property_name})"
        self.append_data = data
        super().__init__(self.message)

    def __str__(self):
        return f'[FlaskSettings] {self.message}'



from . import blueprint
from .settings import SettingClass
from .forms import SettingsForm
from flask import request, current_app
from .error import FormNotFound

@blueprint.get('/')
def ping():
    return 'ping'


@blueprint.post('/s/<setting_key>/set')
def set_value(setting_key):
    setting: SettingClass = SettingClass.get_group(setting_key)
    if not setting:
        return 'Failed'
    setting.set_properties(request.json)
    if current_app.config.get('FLSETT_AUTO_SAVE_SETTINGS'):
        setting.save()
    return 'Success'


@blueprint.post('/f/<form_key>')
def setting_form_enpoint(form_key):
    for subc in SettingsForm.__subclasses__():
        if subc.get_key() == form_key:
            r = subc.on_data(request.json)
            if r is True:
                return 'Success'
            if r is None:
                return 'Unknow State'
            return r
    raise FormNotFound(form=form_key)




from flask import Blueprint, Flask
from werkzeug.local import LocalProxy

blueprint: Blueprint
app: Flask
logger = LocalProxy(lambda: app.logger)
codec_string = 'utf-8'

def create_blueprint(flask_app, name: str = 'settings'):
    global app
    global blueprint
    global logger

    app = flask_app
    _set_defautl_configs(app)

    blueprint = Blueprint(
        name,
        __name__,
        template_folder="templates",
        static_folder="static"
    )
    blueprint.add_app_template_global(name, name='__flasky_settings_name__')

    from flasky_settings import main

    return blueprint

def _set_defautl_configs(app):
    app.config.setdefault('FLSETT_AUTO_SAVE_SETTINGS', False)




from collections import namedtuple

setting_item_tuple = namedtuple(
    'setting_item',
    (
        'key',
        'title',
        'description',      # TODO use in tempate but where
        'real_value'
        )
    )

class SettingItemsList:
    setting_item = namedtuple('setting_item', ('key', 'title', 'description', 'real_value'))
    
    def __init__(self):
        self.items : dict[str:dict] = {}
    
    def get_keys(self):
        return self.items.keys()
    
    def add_item(self, key, title, description=None, real_value=None):
        self.items[key] = setting_item_tuple(key, title, description, real_value)
        
    def key_to_real_value(self, key):
        if key in self.items:
            r = self.items.get(key)
            if r is None:   # TODO to warning or something else
                return key
            return r.real_value
        else:
            raise ValueError(f"Key '{key}' is not settet. Item not exists !")
    
    def loop_items(self):
        for i in self.items.values():
            yield i


from ..error import ValidationError


class SettingsElement:
    template = '/flasky_settings/setting/default.html'
    settings_type = 'default'
    _none_allowed: bool = False
    _default = None
    
    def format_value(self, value):
        return value

    def to_validation(self, v) -> tuple[bool, str]:
        if not self.none_allowed and v is None:
            raise ValidationError(message='"None" is not allowed !', obj=self)

    def _parse_value(self, value):
        return value

    def __init__(self, key, title='', description='', default=None, setting=None, none_allowed: bool | None = None,
                 col='12', **kwargs):
        self.key = key
        self.title = title
        self.description = description

        self.col = col
        self.default = default if default is not None else self._default
        self.setting = setting

        self.none_allowed = none_allowed if none_allowed is not None else self._none_allowed
        self.init(**kwargs)

    def init(self, **kwargs):
        pass





class SettingsForm:
    __key__ = None
    __form__ = []
    __template__ = '/flasky_settings/forms/default.html'
    btn_title = 'Send'

    @classmethod
    def on_data(cls, data):
        return 'Success'

    @classmethod
    def get_form(cls):
        return cls.__form__

    @classmethod
    def loop_elements(cls):
        for element in cls.get_form():
            yield element

    @classmethod
    def get_template(cls):
        return cls.__template__

    @classmethod
    def get_key(cls):
        return cls.__key__ or cls.__name__


import os
import codecs
import json

from ..error import PropertyPermissionError
from .settings_element import SettingsElement
from .. import codec_string, logger

class SettingClass:
    all_settings = {}
    elements: dict[str, SettingsElement] = {}
    settings_path = None

    def __init__(self, key, title):
        self.key = key
        self.title = title

        SettingClass.all_settings[key] = self
        self.load()

    def loop_elements(self):
        for v in self.elements.values():
            yield v

    def set_property(self, k, v):
        if k in self.elements.keys():
            setattr(self, k, v)
        else:
            raise PropertyPermissionError(property_name=k)

    def set_properties(self, d: dict):
        for k, v in d.items():
            self.set_property(k, v)

    def get_property(self, k, default=None) -> any:
        if k in self.elements.keys():
            return self.__getattribute__(k)
        return default

    def get_properties(self) -> dict:
        data = {e: self.get_property(e) for e in self.elements.keys()}
        return data

    def __getitem__(self, item):
        return self.__getattribute__(item)

    def get_value_formattet(self, key):
        se = self.elements.get(key, None)
        v = self.__getattribute__(key)
        v = se.format_value(v)
        return v
    
    # save / loading - configuration

    def generate_file_name(self):
        return f"{self.__class__.__name__}-{self.key}.json"

    def generate_full_file_path(self, filename=None):
        if filename is None:
            filename = self.generate_file_name()
        full_path_filename = os.path.join(self.settings_path, filename)
        return full_path_filename

    def save(self):
        if self.settings_path is None:                      # TODO CHECK IF THIS IST BEST METHODE
            raise Exception("'settings_path' is None")
        # get settings data (for saving)
        data = self.get_properties()
        data_string = json.dumps(data, indent=2)
        # create/save File
        filename = self.generate_file_name()
        full_path_filename = self.generate_full_file_path(filename=filename)
        with codecs.open(full_path_filename, 'w', codec_string) as f:
            f.write(data_string)
            logger.info(f"[{self.__class__.__name__}] Saved Config to '{filename}'")

    def load_default_values(self):
        for e in self.elements.values():
            e: SettingsElement
            setattr(self, e.key, e.default)

    def load_from_file(self, file):
        with codecs.open(file, 'r', codec_string) as f:
            d = json.load(f)
            self.set_properties(d)

    def load(self):
        filename = self.generate_file_name()
        full_path_filename = self.generate_full_file_path(filename=filename)

        if os.path.exists(full_path_filename):
            self.load_from_file(full_path_filename)
            logger.info(f"[{self.__class__.__name__}] Loaded Config from File '{filename}'")
        else:
            logger.info(f"[{self.__class__.__name__}] Load Default Configuration")
            self.load_default_values()

    # class methods

    @classmethod
    def get_group(cls, key, default=None):
        return cls.all_settings.get(key, default)

    @classmethod
    def save_to_path(cls):
        for setting in cls.all_settings.values():
            setting.save()

    @classmethod
    def set_settings_path(cls, path):
        cls.settings_path = path




from regex import R
from ..settings import SettingsElement, SettingClass, SettingItemsList
from ..settings.settings_item import setting_item_tuple
from flasky_settings.error import ValidationError
from datetime import datetime

class SE_Bool(SettingsElement):
    template = '/flasky_settings/setting/bool.html'
    settings_type = 'bool'
    _none_allowed = True

    def to_validation(self, v) -> tuple[bool, str]:
        super().to_validation(v)

        if not isinstance(v, bool):
            raise ValidationError(message=f'ValueType is not correct (Type= {str(type(v))}). Must be Bool !')

    def _parse_value(self, v):
        if isinstance(v, bool):
            return v
        if isinstance(v, str):
            v = v.lower()
            match v:
                case '':
                    return None
                case 'true' | 'on':
                    return True
                case 'false':
                    return False
                case _:
                    raise ValidationError()


class SE_String(SettingsElement):
    template = '/flasky_settings/setting/string.html'
    settings_type = 'string'
    input_type = 'text'
    _default = ''

    def to_validation(self, v) -> tuple[bool, str]:
        super().to_validation(v)
        if not isinstance(v, str):
            raise ValidationError(message=f'[StringElement] ValueType is not correct (Type= {str(type(v))}). Must be String ! ')

    def init(self, placeholder: str = '', **kwargs):            # Todo adding html validation function https://www.w3schools.com/html/tryit.asp?filename=tryhtml5_input_placeholder PATTERN
        self.placeholder = placeholder


class SE_String_Split(SettingsElement):     # TODO Mach das mit einem HTML Template mit add und remove function
    template = '/flasky_settings/setting/string.html'
    settings_type = 'string'
    input_type = 'text'
    _default = []

    def to_validation(self, v) -> tuple[bool, str]:
        super().to_validation(v)
        if not isinstance(v, list):
            raise ValidationError(
                message=f'[StringElement] ValueType is not correct (Type= {str(type(v))}). Must be List ! '
            )

    def _parse_value(self, v: str | list):
        if isinstance(v, str):      # TODO better type checking
            l: list[str] = v.split(';')
            v = list(map(lambda s: s.strip(), l))
        return v

    def init(self, placeholder: str = '', **kwargs):            # Todo adding html validation function https://www.w3schools.com/html/tryit.asp?filename=tryhtml5_input_placeholder PATTERN
        self.placeholder = placeholder


class SE_Text(SettingsElement):
    template = '/flasky_settings/setting/text.html'
    settings_type = 'text'

    def to_validation(self, v) -> tuple[bool, str]:
        super().to_validation(v)
        if not isinstance(v, str):
            raise ValidationError(message=f'[TextElement] ValueType is not correct (Type= {str(type(v))}). Must be String ! ')

    def init(self, rows: int = 3, **kwargs):
        self.rows = rows


class SE_MultiSelect(SettingsElement):
    template = '/flasky_settings/setting/multi_select_as_list.html'
    settings_type = 'multi-select'

    def to_validation(self, v) -> tuple[bool, str]:
        super().to_validation(v)
        if not isinstance(v, list):
            raise ValidationError(message=f'[MultiSelect] ValueType is not correct (Type= {str(type(v))}). Must be List !')

    def init(self, items: list = '', **kwargs):
        self.items = items

    def loop_items(self, setting_group: SettingClass):
        v = setting_group[self.key]
        for i in self.items:
            active = i in v
            yield {
                'name': str(i),
                'active': active
            }


class SE_Select(SettingsElement):
    template = '/flasky_settings/setting/select.html'
    settings_type = 'select'

    def to_validation(self, v) -> tuple[bool, str]:
        super().to_validation(v)

    def init(self, items: list = '', **kwargs):
        self.items = items
        
    def __loop_list(self, active_key):
        for i in self.items:
            active = (i == active_key)
            yield active, {
                'key': str(i),
                'title': str(i),
            } 
            
    def __loop_settings_items(self, active_key):
        self.items: SettingItemsList
        for i in self.items.loop_items():
            active = (i.key == active_key)
            yield active, i

    def loop_items(self, setting_group: SettingClass):
        active_key = setting_group[self.key]
        if isinstance(self.items, list):
            yield from self.__loop_list(active_key)
        elif isinstance(self.items, SettingItemsList):
            yield from self.__loop_settings_items(active_key)
        # TODO raise Unknow Looping Item Type or something else xD
    
    def format_value(self, v):
        if isinstance(self.items, list):
            return v
        elif isinstance(self.items, SettingItemsList):
            return self.items.key_to_real_value(v)
            

class SE_DateTime(SettingsElement):
    template = '/flasky_settings/setting/datetime.html'
    settings_type = 'datetime'
    _default = ''

    def to_validation(self, v) -> tuple[bool, str]:
        super().to_validation(v)
        # TODO date-time parsing




from .settings_property import SettingProperty
from .settings_element import SettingsElement
from .settings_class import SettingClass
from .settings_page import SettingsPage
from .settings_item import SettingItemsList



import os

from setuptools import setup


def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


setup(
    name='flasky-settings',
    version='0.0.2.0',
    url='https://github.com/LordBex/flasky-settings',
    license='',
    author='lordbex',
    author_email='lordibex@protonmail.com',
    description='Flask extension that includes a Settings-Manager in your project',
    long_description=read('README.rst'),
    packages=['flasky_settings', 'flasky_settings.settings', 'flasky_settings.elements', 'flasky_settings.forms'],
    zip_safe=False,
    include_package_data=True,
    platforms='any',
    install_requires=[
        'Flask>=2.2.2'
    ],
    classifiers=[
        'Environment :: Web Environment', 'Intended Audience :: Developers',
        'Operating System :: OS Independent', 'Programming Language :: Python',
        'Programming Language :: Python :: 3.10',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Topic :: Software Development :: Libraries :: Python Modules'
    ])




class SettingsPage:
    pages = {}

    def __init__(self, key, title):
        self.setting_groups = []
        self.title = title
        self.key = key

        SettingsPage.pages[key] = self

    @classmethod
    def get_page(cls, page_key, default=None):
        return cls.pages.get(page_key, default)

    def add_settings_groups(self, *settings_groups):
        for s in settings_groups:
            self.setting_groups.append(s)

    def loop_groups(self):
        for sg in self.setting_groups:
            yield sg
